
def _args_reader(args_parser):  # pragma: no cover
    args = args_parser.add_argument_group('Archive options')

    args.add_argument(
        '-z',
        '--cbz',
        action='store_true',
        help=(
            'Make `*.cbz` archives (for reader).'
        )
    )

    args.add_argument(
        '-R',
        '--rename-pages',
        action='store_true',
        help=(
            'Normalize image filenames. '
            'E.g. from `0_page_1.jpg` to `0001.jpg`.'
        )
    )



from manga_py.provider import Provider
from .helpers.std import Std


class HentaiHereCom(Provider, Std):
    _cdn = 'https://hentaicdn.com/hentai'

    def get_chapter_index(self) -> str:
        chapter = self.chapter
        idx = self.re.search('/m/[^/]+/([^/]+(?:/[^/]+))', chapter)
        return idx.group(1).replace('/', '-')

    def get_content(self):
        url = self.re.search('(/m/[^/]+)', self.get_url())
        url = '{}{}'.format(self.domain, url.group(1))
        return self.http_get(url)

    def get_manga_name(self) -> str:
        selector = 'span.hide[itemscope] span[itemprop="name"]'
        name = self.document_fromstring(self.content, selector)
        if not name:
            selector = '#detail span[itemprop="title"]'
            name = self.document_fromstring(self.content, selector)
        return name[0].text_content().strip()

    def get_chapters(self):
        return self._elements('ul.arf-list > li > a')

    def get_files(self):
        chapter = self.chapter
        content = self.http_get(chapter)
        items = self.re.search(r'_imageList\s*=\s*(\[".+"\])', content).group(1)
        return [self._cdn + i for i in self.json.loads(items)]

    def get_cover(self) -> str:
        return self._cover_from_content('#cover img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = HentaiHereCom



from manga_py.provider import Provider
from manga_py.download_methods import WholeArchiveDownloader
from .helpers.std import Std

class MangaFreakNet(Provider, Std):
    _downloader = WholeArchiveDownloader

    def get_archive_name(self):
        return self.chapter['archive_name']

    def get_chapter_index(self) -> str:
        return self.re.search(r'.+_(\d+)', self.chapter[1]).group(1)

    def get_content(self):
        return self.http_get(self.get_url())

    def get_manga_name(self) -> str:
        return self._get_name('/Manga/([^?/#]+)')

    def get_chapters(self):
        items = self._elements('.manga_series_list td a[download]')
        return [{'archive_name': i.get('download'), 'download_link': i.get('href')} for i in items][::-1]

    def get_files(self):
        pass

    def prepare_cookies(self):
        self.cf_scrape(self.get_url())

    def get_cover(self) -> str:
        return self._cover_from_content('.manga_series_image img')

    def chapter_for_json(self):
        return self.chapter['download_link']


main = MangaFreakNet




from manga_py.provider import Provider
from manga_py.download_methods import WholeArchiveDownloader
from .helpers.std import Std

# Archive downloading example. Without images
class MangaOnlineBiz(Provider, Std):
    _downloader = WholeArchiveDownloader
    chapter_url = ''

    def get_chapter_index(self) -> str:
        url = self.chapter
        idx = self.re.search(r'/download/[^/]+/.+?_(\d+)_(\d+)', url).groups()
        return '{}-{}'.format(*idx)

    def get_content(self):
        return self._get_content('{}/{}.html')

    def get_manga_name(self) -> str:
        return self._get_name(r'\.\w{2,7}/([^/]+)(?:/|\.html)')

    def _after_download(self, idx, _path):
        self._idx = idx + 1

    def get_chapters(self):
        s, c = r'MangaChapter\((.+)\);', self.content
        items = self.json.loads(self.re.search(s, c).group(1))
        return [i.get('downloadUrl') for i in items]

    def get_files(self):
        return []

    def get_cover(self):
        return self._cover_from_content('.item > .image > img')

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.chapter_url


main = MangaOnlineBiz



from manga_py.provider import Provider
from .helpers.std import Std
from urllib import parse


class ComicPunchNet(Provider, Std):

    def get_chapter_index(self) -> str:
        return self.re.search(r'[-/]((?:Annual|Issue|Chapter)-\w+)', self.chapter).group(1)

    def get_content(self):
        return self.http_get(self.get_url())

    def get_manga_name(self) -> str:
        return self.text_content(self.content, '.page-title')

    def get_chapters(self):
        return self._elements('.chapter > a')

    def get_files(self):
        parser = self.html_fromstring(self.chapter + '?q=fullchapter')
        base_url = parser.cssselect('base[href]')[0].get('href')
        return [parse.urljoin(base_url, i) for i in self._images_helper(parser, 'img.picture')]

    def get_cover(self) -> str:
        return self._cover_from_content('.pic .series')


main = ComicPunchNet



from manga_py.provider import Provider
from .helpers.std import Std


class MangaIndoWebId(Provider, Std):

    def get_chapter_index(self) -> str:
        selector = r'-chapter-([^/]+)'
        return self.re.search(selector, self.chapter).group(1)

    def get_content(self):
        return self._get_content('{}/{}/')

    def get_manga_name(self) -> str:
        url = self.get_url()
        pos = url.find('-chapter-')
        if pos > 0:
            item = self.html_fromstring(self.get_url(), 'article[id^="post-"]', 0)
            item = self.re.search(r'category-([^\s]+)', item.get('class')).group(1)
            return item
        return self.re.search(r'\.id/([^/]+)', url).group(1)

    def get_chapters(self):
        return self._elements('.lcp_catlist li > a')

    def get_files(self):
        r = self.http().get_redirect_url
        params = self.chapter, '.entry-content img.aligncenter'
        items = self.html_fromstring(*params)
        return [r(i.get('src')) for i in items]

    def get_cover(self) -> str:
        return self._cover_from_content('#m-cover > img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaIndoWebId



from .gomanga_co import GoMangaCo

from .helpers.std import Std


class HentaiCafe(GoMangaCo, Std):
    _name_re = r'\.cafe/manga(?:/(?:series|read))?/([^/]+)/'
    _content_str = '{}/manga/series/{}/'
    _chapters_selector = '.content .last .x-btn,.element .title a'

    def get_archive_name(self) -> str:
        return 'archive_{:0>2}'.format(self.chapter_id)

    def get_chapter_index(self) -> str:
        return str(self.chapter_id)

    def get_cover(self) -> str:
        return self._cover_from_content('.entry-content img')

    def prepare_cookies(self):
        pass


main = HentaiCafe



from manga_py.provider import Provider
from .helpers.std import Std


class DoujinsCom(Provider, Std):
    img_selector = '#image-container img.doujin'

    def get_archive_name(self) -> str:
        return 'archive'

    def get_chapter_index(self) -> str:
        return '0'

    def get_content(self):
        return self._get_content('{}/gallery/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/gallery/([^/]+)')

    def get_chapters(self):
        return [b'']

    def get_files(self):
        items = self.document_fromstring(self.content, self.img_selector)
        return [i.get('data-file').replace('&amp;', '&') for i in items]

    def get_cover(self) -> str:
        return self._cover_from_content(self.img_selector)

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.get_url()


main = DoujinsCom



from abc import ABCMeta, abstractmethod
from pathlib import Path
from sys import platform

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver as BrowserDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.expected_conditions import (
    presence_of_all_elements_located as locate,
    presence_of_element_located as locate_one
)
from selenium.webdriver.support.ui import WebDriverWait
from typing import List
from time import sleep

from manga_py.fs import get_util_home_path

class UnsupportedOsException(Exception):
    def __init__(self, os=platform, *args, **kwargs):
        super().__init__(os, *args, **kwargs)


def _is_win():
    return platform.startswith('win32') or platform.startswith('cygwin')


def _is_mac():
    return platform.startswith('darwin')


def _is_linux():
    return platform.startswith('linux')


def _supported() -> bool:
    return _is_win() or _is_mac() or _is_linux()


class WebDriver(metaclass=ABCMeta):
    driver_version = None
    home_path = None
    _re = None
    _driver = None
    _default_timeout = 10
    __visible = True
    __freq = 1.0

    def __init__(self, version: str, _re):
        self.driver_version = version
        self._re = _re
        self.home_path = Path(get_util_home_path())

    @property
    def visible(self) -> bool:
        return self.__visible

    @visible.setter
    def visible(self, visible: bool):
        self.__visible = visible

    @staticmethod
    @abstractmethod
    def driver_archive() -> str:
        raise NotImplementedError

    @abstractmethod
    def driver_path(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def download_driver(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _make_driver(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_available_versions(self) -> set:
        raise NotImplementedError

    def _version(self, version: str) -> str:
        return self._re.search(version).group(1)

    @property
    def initialized(self) -> bool:
        return self._driver is not None

    def init_driver(self):
        if not self.initialized:
            self._make_driver()
        return self

    @property
    def freq(self) -> float:
        return float(self.__freq)

    @freq.setter
    def freq(self, freq: float):
        self.__freq = freq

    @property
    def driver(self) -> BrowserDriver:
        if not self.initialized:
            self.init_driver()
        return self._driver

    def get(self, url):
        self.driver.get(url)

    @property
    def valid_url(self):
        if self.initialized:
            url = self._driver.current_url
            if url and url.startswith('http'):
                return True
        return False

    @property
    def _wait(self):
        return WebDriverWait(self._driver, self._default_timeout, poll_frequency=self.freq)

    def find_element(self, selector: str, by=By.CSS_SELECTOR) -> List[WebElement]:
        sleep(1)
        return self._wait.until(locate_one((by, selector)))

    def find_all_elements(self, selector: str, by=By.CSS_SELECTOR) -> List[WebElement]:
        sleep(1)
        self._wait.until(locate_one((by, selector)))  # wait for first element
        sleep(1)
        return self._wait.until(locate((by, selector)))

    def add_cookie(self, key, value, **kwargs):
        _cookie = {
            'name': key,
            'value': value,
            'expiry': 1898789118,
            'secure': False,
            'path': '/',
            'httpOnly': False,
        }

        _cookie.update(kwargs)

        self._driver.add_cookie(_cookie)

    def close(self):
        if self.initialized:
            self._driver.quit()


__all__ = ['WebDriver', '_is_win', '_is_mac', '_is_linux', '_supported', 'UnsupportedOsException']



from manga_py.provider import Provider
from .helpers.std import Std


class MangaOnlineCom(Provider, Std):
    __local_storage = None

    def __init_storage(self):
        if not self.__local_storage:
            self.__local_storage = {}

    def get_chapter_index(self) -> str:
        self.__init_storage()
        idx_reg = r'/\d+.+-(\d+).+?-(\d+).*?html'
        idx = self.re.search(idx_reg, self.chapter).groups()
        if not idx:
            idx_reg = r'/\d+.+-(\d+).+?html'
            idx = (self.re.search(idx_reg, self.chapter).group(1), 0)
        return '{:0>3}-{:0>3}'.format(*idx)

    def get_content(self):
        return ['0']

    def get_manga_name(self) -> str:
        self.__init_storage()
        if not self.__local_storage.get('chapters', False):
            self.__local_storage['chapters'] = self.get_chapters()
        if len(self.__local_storage['chapters']):
            return self.re.search(r'/manga/(.+)/.+\.html', self.__local_storage['chapters'][0]).group(1)
        raise AttributeError()

    def _get_chapters_cmanga(self):
        s = '#dle-content > div > a[href*="/manga/"]'
        return self.html_fromstring(self.get_url(), s)[::-1]

    def _get_chapters_manga(self):
        s = '.fullstory_main select.selectmanga option'
        items = self.html_fromstring(self.get_url(), s)
        return [i.get('value') for i in items[::-1]]

    def get_chapters(self):
        self.__init_storage()
        if self.re.search('/cmanga/', self.get_url()):
            return self._get_chapters_cmanga()
        if self.re.search(r'/manga/[^/]+/\d+-', self.get_url()):
            return self._get_chapters_manga()
        return []

    @staticmethod
    def _get_pages_count(parser):
        _len = len(parser.cssselect('#pages_all a'))
        return _len + 1 if _len else 0

    def get_files(self):
        chapter = self.chapter
        parser = self.html_fromstring(chapter, '.main_body', 0)
        pages = self._get_pages_count(parser)
        images = []
        idx = self.re.search(r'/manga/[^/]+/(\d+)', chapter).group(1)
        for n in range(pages):
            url = '{}/engine/ajax/sof_fullstory.php?id={}&page={}'.format(self.domain, idx, n + 1)
            parser = self.html_fromstring(url)[0]
            images += self._images_helper(parser, 'img')
        return images

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaOnlineCom



from manga_py.provider import Provider
from .helpers.std import Std


class CycomiCom(Provider, Std):
    @staticmethod
    def remove_not_ascii(value):
        return value

    def get_chapter_index(self) -> str:
        return self.chapter[1]

    def get_content(self):
        idx = self.re.search(
            r'/title/(\d+)',
            self.get_url()
        ).group(1)

        url = '{}/fw/cycomibrowser/chapter/title/{}'.format(
            self.domain,
            idx
        )

        return self.http_get(url)

    def get_manga_name(self) -> str:
        return self.text_content(self.content, '.title-texts h3')

    def get_chapters(self):
        selector = 'a.chapter-item:not(.is-preread)'
        items = []
        n = self.http().normalize_uri
        for el in self._elements(selector, self.content):
            title = el.cssselect('p.chapter-title')[0]
            title = title.text_content().strip(' \n\r\t\0')
            episode_id = self.re.sub(r'.+pages/(.+)', r'\1', n(el.get('href')))
            title = episode_id + '_' + title
            items.append((n(el.get('href')), title))
        return items

    def get_files(self):
        content = self.http_get(self.chapter[0])
        parser = self.document_fromstring(content)
        selector = '.comic-image'
        return self._images_helper(parser, selector)

    def get_cover(self) -> str:
        return self._cover_from_content('.title-image-container img')

    def chapter_for_json(self):
        return self.chapter[1]

    def book_meta(self) -> dict:
        # todo meta
        pass


main = CycomiCom



from .base_lib import BaseLib


class HentaiNexusComCrypt(BaseLib):
    def __init__(self):
        super().__init__()
        self._this = []
        self._pads = {}

    def decode(self, enc):
        buff = self.base64decode(enc)

        i = 0
        file = 0
        key = 0
        _hash = 0
        _input = []

        self._step_0()
        c = self._step_1(buff)
        index = self._step_3(buff, self._step_2())

        _next = self._this[c]

        for offset in range(64, len(buff)):
            i = (i + _next) % 256
            key = (file + index[(key + index[i]) % 256]) % 256
            file = (file + i + index[i]) % 256

            index[i], index[key] = index[key], index[i]

            _hash = index[(key + index[(i + index[(_hash + file) % 256]) % 256]) % 256]

            _input.append(chr(buff[offset] ^ _hash))

        return ''.join(_input)

    def _step_0(self):
        index = 2
        while len(self._this) < 16:
            try:
                _ = self._pads[index]
            except KeyError:
                self._this.append(index)
                week_index = index << 1

                while week_index <= 256:
                    self._pads[week_index] = True

                    week_index = week_index + index

            index += 1

    def _step_1(self, buff) -> int:
        c = 0

        for index in range(0, 64):
            c = c ^ buff[index]

            for week_index in range(0, 8):
                c = c >> 1 ^ 12 if c & 1 else c >> 1

        return c & 7

    def _step_2(self) -> list:
        return list(range(0, 256))

    def _step_3(self, buff, index: list) -> list:
        key = 0
        for i in range(0, 256):
            key = (key + index[i] + buff[i % 64]) % 256
            index[i], index[key] = index[key], index[i]

        return index





from PIL import Image


# DO NOT SEE HERE! IT WORKED!


class MatrixSunday:
    __image_src = None
    __image_dst = None

    def de_scramble(self, path_src: str, path_dst: str, data: list):
        self.__image_src = Image.open(path_src, 'r')
        self.__process(data)
        self.__image_dst.save(path_dst)
        self.__image_src.close()
        self.__image_dst.close()

    def __process(self, data: list):
        size_src = self.__image_src.size

        self.__image_dst = Image.new(self.__image_src.mode, size_src)

        for i in data:
            x, y = i['srcX'] + i['width'], i['srcY'] + i['height']
            dx, dy = i['destX'] + i['width'], i['destY'] + i['height']

            c1 = i['srcX'] < size_src[0]
            c2 = i['srcX'] + i['width'] >= 0
            c3 = i['srcY'] < size_src[1]
            c4 = i['srcY'] + i['height'] >= 0
            if c1 and c2 and c3 and c4:
                region = self.__image_src.crop((i['destX'], i['destY'], dx, dy))
                self.__image_dst.paste(region, (i['srcX'], i['srcY'], x, y))


class SundayWebryCom:  # pragma: no cover
    _result = None

    def solve_by_img(self, src: str, element_width: int, element_height: int, n: int):
        img = Image.open(src)
        sizes = img.size
        img.close()
        return self.solve(*sizes, element_width, element_height, n)

    def solve(self, width: int, height: int, element_width: int, element_height: int, n: int):
        e = width
        t = height
        r = element_width
        i = element_height

        y = int(e / r)
        g = int(t / i)
        f = e % r
        b = t % i
        self._result = []

        s = y - 43 * n % y
        if s % y == 0:
            s = (y - 4) % y

        a = g - 47 * n % g
        if a % g == 0:
            a = (g - 4) % g
        if 0 == a:
            a = g - 1

        self.def1(f, b, s, r, a, i)

        self.def2(y, i, n, a, s, f, r, g, b)

        if f > 0:
            self.def3(g, n, s, a, y, b, i, r, f)

        self.def4(y, g, n, r, f, s, a, i, b)

        return self._result

    def def1(self, f, b, s, r, a, i):
        if f > 0 and b > 0:
            o = s * r
            u = a * i
            self._result.append({
                'srcX': o,
                'srcY': u,
                'destX': o,
                'destY': u,
                'width': f,
                'height': b,
                # 'debug': 1
            })

    def def2(self, y, i, n, a, s, f, r, g, b):
        for l in range(y):
            d = self._calc_x_x(l, y, n)
            h = self._calc_y_x(d, s, a, g, n)
            c = self._calc_pos_rest(d, s, f, r)
            p = h * i
            o = self._calc_pos_rest(l, s, f, r)
            u = a * i
            self._result.append({
                'srcX': o,
                'srcY': u,
                'destX': c,
                'destY': p,
                'width': r,
                'height': b,
                # 'debug': 2
            })

    def def3(self, g, n, s, a, y, b, i, r, f):
        for m in range(g):
            h = self._calc_y_y(m, g, n)
            d = self._calc_x_y(h, s, a, y, n)
            p = self._calc_pos_rest(h, a, b, i)
            u = self._calc_pos_rest(m, a, b, i)
            self._result.append({
                'srcX': s * r,
                'srcY': u,
                'destX': d * r,
                'destY': p,
                'width': f,
                'height': i,
                # 'debug': 3
            })

    def def4(self, y, g, n, r, f, s, a, i, b):
        for l in range(y):
            for m in range(g):
                d = (l + 29 * n + 31 * m) % y
                h = (m + 37 * n + 41 * d) % g
                c = d * r + (f if d >= self._calc_x_y(h, s, a, y, n) else 0)
                p = h * i + (b if h >= self._calc_y_x(d, s, a, g, n) else 0)
                o = l * r + (f if l >= s else 0)
                u = m * i + (b if m >= a else 0)
                self._result.append({
                    'srcX': o,
                    'srcY': u,
                    'destX': c,
                    'destY': p,
                    'width': r,
                    'height': i,
                    # 'debug': 4
                })

    @staticmethod
    def _calc_pos_rest(e, t, r, i):
        m = 0
        if e >= t:
            m = r
        return e * i + m

    @staticmethod
    def _calc_x_x(e, t, r):
        return (e + 61 * r) % t

    @staticmethod
    def _calc_x_y(e, t, r, i, n):
        o = (n % 2 == 1)
        if (e < r and o) or (e >= r and not o):
            a = i - t
            s = t
        else:
            a = t
            s = 0
        return (e + 67 * n + t + 71) % a + s

    @staticmethod
    def _calc_y_x(e, t, r, i, n):
        o = (n % 2 == 1)
        if (e < t and o) or (e >= t and not o):
            a = r
            s = 0
        else:
            a = i - r
            s = r
        return (e + 53 * n + 59 * r) % a + s

    @staticmethod
    def _calc_y_y(e, t, r):
        return (e + 73 * r) % t



from manga_py.provider import Provider
from .helpers.std import Std


class WebAceJp(Provider, Std):
    @staticmethod
    def remove_not_ascii(value):
        return value

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index().split('-', 2)
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        idx = self.re.search(
            r'第(.+?)?話(?:-(.+?))?',
            self.chapter[1]
        )
        if not idx:
            return self.chapter[1]
        return self._join_groups(idx.groups())

    def __url(self):
        return '{}/youngaceup/contents/{}/'.format(
            self.domain,
            self.__idx()
        )

    def __idx(self):
        return self.re.search(
            r'/contents/(\d+)',
            self.get_url()
        ).group(1)

    def get_content(self):
        return self.http_get(self.__url())

    def get_manga_name(self) -> str:
        return self.text_content(self.content, '.credit h1')

    def get_chapters(self):
        content = self.http_get(self.__url() + 'episode/')
        selector = '.media:not(.yudo) > a.navigate-right'
        items = []
        n = self.http().normalize_uri
        for el in self._elements(selector, content):
            title = el.cssselect('.media-body p')[0]
            title = title.text_content().strip(' \n\r\t\0')
            items.append((n(el.get('href')), title))
        return items

    def get_files(self):
        n = self.http().normalize_uri
        with self.http().get(self.chapter[0] + '/json/') as resp:
            items = resp.json()
        return [n(i) for i in items]

    def get_cover(self) -> str:
        return self._cover_from_content('#sakuhin-info img')

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.chapter[0]


main = WebAceJp



from .gomanga_co import GoMangaCo


class NoraNoFansubCom(GoMangaCo):
    _name_re = r'\.\w{2,7}/(?:lector/)?(?:series/|read/)?([^/]+)/'
    _content_str = '{}/{}/'
    _chapters_selector = '.entry-content td a[href]'

    def get_chapters(self):
        return super().get_chapters()[::-1]

    def get_cover(self) -> str:
        return self._cover_from_content('.entry-content img.size-full')


main = NoraNoFansubCom



from manga_py.crypt.base_lib import BaseLib
from manga_py.provider import Provider
from .helpers.std import Std


class MangaFoxMe(Provider, Std):

    def get_archive_name(self) -> str:
        groups = self._ch_parser()
        ch = groups[1].replace('.', '-')
        vol = ['0']
        if groups[0]:
            vol = [groups[0]]
        return self.normal_arc_name({'vol': vol, 'ch': ch})

    def _ch_parser(self):
        selector = r'/manga/[^/]+/(?:v([^/]+)/)?c([^/]+)/'
        groups = self.re.search(selector, self.chapter).groups()
        return groups

    def get_chapter_index(self) -> str:
        groups = self._ch_parser()
        idx = groups[1].replace('.', '-')
        if not ~idx.find('-'):
            idx = idx + '-0'
        if groups[0]:
            return '{}-{}'.format(idx, groups[0])
        return idx

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)/?')

    def get_chapters(self):
        return self._elements('[id^="list-"] a[href]')

    def _get_links(self, content):
        js = self.re.search(r'eval\((function\b.+)\((\'[\w ].+)\)\)', content).groups()
        return BaseLib.exec_js('m = ' + js[0], 'm(' + js[1] + ')')

    def _one_link_helper(self, content, page):
        cid = self.re.search(r'chapterid\s*=\s*(\d+)', content).group(1)
        base_url = self.chapter[0:self.chapter.rfind('/')]
        links = self._get_links(content)
        key = ''.join(self.re.findall(r'\'(\w)\'', links))
        return self.http_get('{}/chapterfun.ashx?cid={}&page={}&key={}'.format(
            base_url,
            cid,
            page,
            key
        ))

    def _parse_links(self, data):
        base_path = self.re.search(r'pix="(.+?)"', data).group(1)
        images = self.re.findall(r'"(/\w.+?)"', data)
        return [base_path + i for i in images]

    def _get_links_page_to_page(self, content):
        last_page = self.document_fromstring(content, '.pager-list-left > span > a:nth-last-child(2)', 0)
        links = []
        for i in range(0, int(int(last_page.get('data-page')) / 2 + .5)):
            data = self._one_link_helper(content, (i * 2) + 1)
            links += self._parse_links(self._get_links(data))
        return links

    def get_files(self):
        content = self.http_get(self.chapter)
        links = self._get_links(content)

        n = self.http().normalize_uri

        if ~links.find('key='):
            # chapters data example: http://fanfox.net/manga/the_hero_is_overpowered_but_overly_cautious/c001/chapterfun.ashx?cid=567602&page=6&key=6b5367d728d445a8
            return self._get_links_page_to_page(content)

        if ~links.find('token='):
            links_array = self.re.search(r'(\[.+?\])', links)
            links_array = links_array.group(1).replace('\'', '"')
            links_data = self.json.loads(links_array)
            return [n(i) for i in links_data]

        data = self.re.search(r'\w=(\[.+\])', links).group(1)
        data = self.json.loads(data.replace("'", '"'))
        return [n(i) for i in data]

    def get_cover(self):
        return self._cover_from_content('img.detail-info-cover-img')

    def book_meta(self) -> dict:
        # todo meta
        pass

    def prepare_cookies(self):
        self.http().cookies['isAdult'] = '1'


main = MangaFoxMe



from manga_py.provider import Provider
from .helpers.std import Std
import requests


class RawDevArtComOld(Provider, Std):
    _chapter_selector = r'/chapter/[^\d]+(\d+(?:\.\d+)?)'

    def get_chapter_index(self) -> str:
        idx = self.re.search(self._chapter_selector, self.chapter)
        return '-'.join(idx.group(1).split('.'))

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        items = self._elements('.wp-manga-chapter > a')

        if len(items) == 0:  # example: https://mangasushi.net
            manga_id = self._elements('#manga-chapters-holder')[0].get('data-id')
            items_content = requests.post('{}/wp-admin/admin-ajax.php'.format(self.domain), data={
                'action': 'manga_get_chapters',
                'manga': manga_id,
            }).text
            items = self._elements('.wp-manga-chapter > a', items_content)


        return items

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        return self._images_helper(parser, '.page-break img.wp-manga-chapter-img')

    def get_cover(self) -> str:
        return self._cover_from_content('.summary_image img.img-responsive')


main = RawDevArtComOld



from manga_py.provider import Provider
from .helpers.std import Std
from manga_py.crypt.base_lib import BaseLib


class MangaHereCc(Provider, Std):

    def get_chapter_index(self) -> str:
        chapter = self.chapter
        selector = r'/manga/[^/]+/[^\d]+(\d+)/[^\d]+(\d+)'
        idx = self.re.search(selector, chapter)
        if idx:
            return '-'.join(idx.groups())
        selector = r'/manga/[^/]+/[^\d]+(\d+)'
        return self.re.search(selector, chapter).group(1)

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        return self._elements('.detail-main-list a')

    def get_files(self):
        content = self.http_get(self.chapter)
        chapter_id = self.re.search(r'(?:var)?\s+chapterid\s*=\s*(\d+);', content).group(1)
        guid_key = self.guid_key(content)

        pages = int(self.document_fromstring(content, '.pager-list-left span > a', -2).get('data-page'))
        chapter_url = self.re.search(r'^(.+/c\d+(?:\.\d+))/', self.chapter).group(1)

        n = self.http().normalize_uri
        return [n(i) for i in self._parse_images(pages, chapter_url, chapter_id, guid_key)]

    def guid_key(self, content):
        js = self.re.search(r'>\s*eval(\(function.+\))\s*<', content).group(1)
        key = BaseLib.exec_js('var j=' + js, 'j').split(';')
        return BaseLib.exec_js(key[0], 'guidkey')

    def _parse_images(self, pages, chapter_url, chapter_id, guid_key):
        images = []
        for page in range(0, pages + 1, 2):
            images_content = self.http_get('{}/chapterfun.ashx?cid={}&page={}&key={}'.format(
                chapter_url,
                chapter_id,
                page,
                guid_key,
            ))
            js = self.re.search(r'eval\((.+)\)', images_content).group(1)
            img = BaseLib.exec_js(BaseLib.exec_js('var _p = ' + js, '_p'), 'd')
            for i in img:
                if i not in images:
                    images.append(i)
        return images

    def get_cover(self):
        return self._cover_from_content('.detail-info-cover-img')

    def prepare_cookies(self):
        self._base_cookies()
        self.http().cookies['isAdult'] = '1'


main = MangaHereCc



from manga_py.provider import Provider
from .helpers.std import Std


class MangaPandaCom(Provider, Std):
    _cdn = 'https://img.mghubcdn.com/file/imghub'
    _api_url = 'https://api.mghubcdn.com/graphql'

    def get_chapter_index(self) -> str:
        return self.chapter_for_json()

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name(r'\.\w{2,7}/manga/([^/]+)')

    def get_chapters(self):
        # curl 'https://api.mghubcdn.com/graphql' -H 'Content-Type: application/json' --data-raw '{"query": "{latestPopular(x:mr02){id}manga(x:mr02,slug:\"mahou-tsukai-no-yome\"){id,rank,title,slug,status,image,latestChapter,author,artist,genres,description,alternativeTitle,mainSlug,isYaoi,isPorn,isSoftPorn,unauthFile,noCoverAd,isLicensed,createdDate,updatedDate,chapters{id,number,title,slug,date}}}"}'

        keys = [
            'id', 'rank', 'title', 'slug', 'status', 'chapters'
        ]

        data = self._api(
            r'{manga(x:mr02,slug:"%(name)s"){%(keys)s{id,number,title,slug,date}}}'
            % {'name': self.manga_name, 'keys': ','.join(keys)}
        )

        chapters = data.get('data', {}).get('manga', {}).get('chapters', [])  # type: list

        return chapters[::-1]

    def get_files(self):
        data = self._api(
            r'{chapter(x:mr02,slug:"%(name)s",number:%(number)f){id,title,mangaID,number,slug,date,pages}}'
            % {'name': self.manga_name, 'number': self.chapter['number']}
        )

        images = data.get('data', {}).get('chapter', {}).get('pages', [])

        if type(images) == str:
            images = self.json.loads(images)  # type: list

        return list(map(lambda x: '{}/{}'.format(self._cdn, images[x]), images))

    def get_cover(self):
        return self._cover_from_content('.manga-thumb')

    def chapter_for_json(self) -> str:
        number = str(self.chapter['number']).replace('.', '-')
        slug = self.chapter['slug']

        return '{}_{}'.format(number, slug)

    def _api(self, query: str) -> dict:
        response = self.http().requests(self._api_url, method='post', data={
            "query": query,
        }, headers={
            'Accept': 'application/json',
        })

        return response.json()


main = MangaPandaCom



import json
from os import chmod
from pathlib import Path
from re import compile
from tarfile import open as tar_open
from zipfile import ZipFile

from requests import get
from selenium.webdriver import Firefox
from selenium.webdriver.firefox.options import Options

from .web_driver import WebDriver, _is_win, _is_mac, _is_linux, UnsupportedOsException


class FirefoxDriver(WebDriver):
    url_api = 'https://api.github.com/repos/mozilla/geckodriver/releases'
    url_driver = 'https://github.com/mozilla/geckodriver/releases/download/{ver}/{arc}'
    _arc = 'geckodriver-{ver}-{os}.{ext}'

    def __init__(self, version: str = None):
        super().__init__(('v%s' % (version or '0.26.0').lstrip('v')), compile(r'(.+)/'))

    def driver_archive(self) -> str:
        if _is_linux():
            return self._arc.format(ver=self.driver_version, os='linux64', ext='tar.gz')
        if _is_mac():
            return self._arc.format(ver=self.driver_version, os='macos', ext='tar.gz')
        if _is_win():
            return self._arc.format(ver=self.driver_version, os='win64', ext='zip')
        raise UnsupportedOsException()

    def driver_path(self) -> Path:
        _driver = 'geckodriver'
        if _is_win():
            _driver += '.exe'
        return self.home_path.joinpath(_driver)

    def download_driver(self) -> None:
        driver = 'driver'
        if _is_win():
            driver += '.zip'
        else:
            driver += '.tgz'
        path = str(self.home_path.joinpath(driver))

        with open(path, 'wb') as _driver:
            _driver.write(get(self.url_driver.format(
                ver=self.driver_version,
                arc=self.driver_archive()
            )).content)
            _driver.close()
        if _is_win():
            with ZipFile(path) as file:
                _path = str(self.driver_path().parent)
                file.extractall(_path)
        else:
            _path = str(self.driver_path().parent)
            tar_open(path).extractall(_path)

    def _make_driver(self) -> None:
        _path = self.driver_path()
        driver_path = str(_path)
        if not _path.is_file():
            self.download_driver()
        _is_win() or chmod(driver_path, 0o755)
        options = Options()
        options.set_preference("media.volume_scale", "0.0")
        options.headless = not self.visible
        self._driver = Firefox(
            executable_path=driver_path,
            options=options,
        )

    def get_available_versions(self) -> set:
        content = json.loads(get(self.url_api).text)
        return {i['tag_name'][1:] for i in content if 'master' == i['target_commitish']}


__all__ = ['FirefoxDriver']



from manga_py.provider import Provider
from .helpers.std import Std


class NeuMangaTv(Provider, Std):

    def get_chapter_index(self) -> str:
        chapter = self.chapter
        idx = self.re.search(r'/manga/[^/]+/(\d+(?:\+\d+))', chapter).group(1)
        return '-'.join(idx.split('+'))

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        return self._elements('#scans .item-content a')

    def get_files(self):
        img_selector = '.imagechap'
        parser = self.html_fromstring(self.chapter)
        pages = self._first_select_options(parser, '.readnav select.page')
        images = self._images_helper(parser, img_selector)
        for i in pages:
            url = i.get('value').replace('//', '/').replace(':/', '://')
            images += self._images_helper(self.html_fromstring(url), img_selector)
        return images

    def get_cover(self) -> str:
        return self._cover_from_content('.info img.imagemg')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = NeuMangaTv



from .base_lib import BaseLib


class MangaRockComCrypt(BaseLib):
    def decrypt(self, string):
        if isinstance(string, str):
            string = string.encode()
        n = len(string) + 7

        tmp = self.pack_auto([82, 73, 70, 70])

        tmp += self.pack_auto([
            n & 0xff,
            (n >> 8) & 0xff,
            (n >> 16) & 0xff,
            (n >> 24) & 0xff,
        ])

        tmp += self.pack_auto([87, 69, 66, 80, 86, 80, 56])
        for i in range(len(string)):
            tmp += self.pack('1B', string[i] ^ 101)
        return tmp



from manga_py.crypt import Puzzle
from manga_py.fs import get_temp_path, rename
from manga_py.provider import Provider


class TonariNoYjJp:
    provider = None
    div_num = 4
    multiply = 8
    matrix = None
    temp_path = None

    def __init__(self, provider: Provider):
        self.provider = provider
        self.temp_path = get_temp_path('__image_matrix{}.png')
        matrix = {}
        for i in range(self.div_num * self.div_num):
            matrix[i] = (i % self.div_num) * self.div_num + int(i / self.div_num)
        self.matrix = matrix

    def _chapter_api_content(self, idx) -> dict:
        api = '{}/api/viewer/readable_products?current_readable_product_id={}&' \
              'number_since=99&number_until=-1&read_more_num=100&type=episode'
        content = self.provider.http_get(api.format(self.provider.domain, idx))
        if content[0] == '{':
            return self.provider.json.loads(content)
        return {}

    def _check_need_next_chapter(self, next_url):
        if next_url:
            test = self.provider.re.search('number_since=(\d+)', next_url).group(1)
            if int(test) > 1:
                return True
        return False

    def get_chapters(self, idx) -> list:
        content = self._chapter_api_content(idx)
        items = self.provider.document_fromstring(content.get('html', '<html></html>'), '.series-episode-list-thumb')
        need_more = self._check_need_next_chapter(content.get('nextUrl', None))
        if need_more:
            items += self.get_chapters(content.get('nextUrl'))
        re = self.provider.re.compile(r'/episode-thumbnail/(\d+)')
        return [re.search(i.get('src')).group(1) for i in items]

    def solve_image(self, path, idx):
        try:
            solver = Puzzle(self.div_num, self.div_num, self.matrix, self.multiply)
            solver.need_copy_orig = True
            _ = self.temp_path.format(idx)
            solver.de_scramble(path, _)
            rename(_, path)
        except Exception:
            pass



# from manga_py.fs import dirname, path_join, get_temp_path, rename
from manga_py.provider import Provider
from manga_py.download_methods import WholeArchiveDownloader
from .helpers.std import Std

class MangaChanMe(Provider, Std):
    _downloader = WholeArchiveDownloader

    def get_chapter_index(self) -> str:
        return str(self.chapter_id)

    def get_content(self):
        pass

    def get_manga_name(self) -> str:
        name = r'\.\w{2,7}/[^/]+/\d+-(.+)\.html'
        return self._get_name(name)

    #def loop_chapters(self):
        # Could not test, website does not seem to exist anymore...
        # previously there was a normalize_uri on every chapters,
        # but it should already be handled in __normalize_chapters in Base class

    def get_chapters(self):
        selector = r'\.\w{2,7}/[^/]+/(\d+-.+\.html)'
        url = self._get_name(selector)
        url = '{}/download/{}'.format(self.domain, url)
        return self.html_fromstring(url, 'table#download_table tr td + td > a')[::-1]

    def get_files(self):
        return []

    def get_cover(self):
        selector = r'\.\w{2,7}/[^/]+/(\d+-.+\.html)'
        url = '{}/manga/{}'.format(self.domain, self._get_name(selector))
        img = self._elements('#cover', self.http_get(url))
        if img and len(img):
            return img[0].get('src')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaChanMe



from manga_py.provider import Provider
from .helpers.std import Std


class HocVienTruyenTranhCom(Provider, Std):

    def get_chapter_index(self) -> str:
        idx = self.re.search(r'/chapter/(\d+)', self.chapter)
        return '{}-{}'.format(self.chapter_id, idx.group(1))

    def _test_main_url(self, url):
        if self.re.search('/chapter/', url):
            url = self.html_fromstring(url, '#subNavi a', 0).get('href')
        return url

    def get_content(self):
        url = self._test_main_url(self.get_url())
        return self.http_get(self.http().normalize_uri(url))

    def get_manga_name(self) -> str:
        url = self._test_main_url(self.get_url())
        return self.re.search('/manga/[^/]+/([^/]+)', url).group(1)

    def get_chapters(self):
        return self._elements('.table-scroll table.table td > a')

    def get_files(self):
        selector = '.manga-container img.page'
        items = self.html_fromstring(self.chapter, selector)
        return [i.get('src') for i in items]

    def get_cover(self):
        return self._cover_from_content('.__info-container .__image img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = HocVienTruyenTranhCom



from manga_py.provider import Provider
from .helpers.std import Std


class MangaOnlineComBr(Provider, Std):

    def get_chapter_index(self) -> str:
        selector = r'\.\w{2,7}/[^/]+/[^/]+/([^/]+)'
        return self.re.search(selector, self.chapter).group(1)

    def get_content(self):
        return self._get_content('{}/{}/')

    def get_manga_name(self) -> str:
        return self._get_name(r'\.\w{2,7}/([^/]+)')

    def get_chapters(self):
        return self._elements('#volumes-capitulos span > a')

    @staticmethod
    def _get_pages_count(parser):
        pages = parser.cssselect('select.pagina-capitulo')
        if pages:
            return len(pages[0].cssselect('option + option'))
        return 0

    def get_files(self):
        img_selector = '#imgPadraoVisualizacao img'
        url = '{}/capitulo.php?act=getImg&anime={}&capitulo={}&src={}&view=1'
        params = (
            self.domain,
            self.manga_name,
            self.get_chapter_index()
        )
        parser = self.html_fromstring(url.format(*params, 1))
        images = self._images_helper(parser, img_selector)
        pages = self._get_pages_count(parser)
        if pages:
            for i in range(int(pages / 2)):
                parser = self.html_fromstring(url.format(*params, ((i + 1) * 2 + 1)))
                images += self._images_helper(parser, img_selector)

        return images

    def get_cover(self) -> str:
        return self._cover_from_content('.image > img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaOnlineComBr



import time
from json import loads
from logging import error, info
from sys import stderr

from requests import request
from requests.cookies import cookiejar_from_dict

from manga_py import meta
from manga_py.crypt.viz_com import solve
from manga_py.fs import get_util_home_path, path_join, is_file, unlink, file_size
from manga_py.provider import Provider
from .helpers.std import Std


class VizDownloader:

    def viz_downloader(self, file_name, url, method):
        with open(file_name, 'wb') as out_file:
            response = request(
                method, url, timeout=20,
                allow_redirects=True,
                headers={
                    'Referer': 'https://www.viz.com/shonenjump/',
                    'User-Agent': self.http().user_agent,
                },
                cookies=cookiejar_from_dict(self.http().cookies),
            )

            if 200 >= response.status_code < 300:
                out_file.write(response.content)
                response.close()
                out_file.close()


class VizCom(Provider, Std, VizDownloader):
    cookie_file = None
    __cookies = {}
    __has_auth = False
    _continue = True
    _metadata = None

    def get_chapter_index(self) -> str:
        # return str(self.chapter_id)
        # todo: need tests
        idx = str(self.chapter_id)
        try:
            re = self.re.compile(r'-chapter-(\d+)/')
            idx = re.search(self.chapter).group(1)
        except AttributeError:
            self.log('manga-py can not get the number of the chapter!\nurl: {}'.format(self.chapter), file=stderr)
            self.log(' Please, report this error\n {}{}\n\n'.format(
                meta.repo_url, '/issues/new?template=bug_report.md'
            ), file=stderr)
        info('Chapter idx: {}'.format(idx))
        return idx

    def get_content(self):
        return self._get_content('{}/shonenjump/chapters/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/chapters/([^/]+)')

    def get_chapters(self):
        chapters = []
        for chapter in self._elements('a.o_chapter-container[href*="/chapter/"]'):
            url = chapter.get('href')
            if url not in chapters:
                chapters.append(url)

        # Paid chapters are dynamically loaded so we need to take a different approach.
        re = self.re.compile(r'targetUrl:\'(.*)\',targetTitle')
        for chapter in self._elements('a.o_chapter-container[onclick*="/chapter/"]'):
            url = re.search(chapter.get('onclick')).group(1)
            if url not in chapters:
                chapters.append(url)

        info('Chapters count: %d' % len(chapters))

        return chapters

    def get_files(self):
        _proxies = self.http().proxies.copy()
        self.http().proxies = dict([])

        self._continue = True
        ch = self.chapter

        params = [
            'device_id=3',
            'manga_id={}'.format(self.re.search(r'/chapter/(\d+)', ch).group(1)),
            'metadata=1',
        ]
        url = '{}/manga/get_manga_url?'.format(self.domain) + '&'.join(params)
        self.log(self.http_get(self.http().normalize_uri(url)))
        __url = self.http_get(self.http().normalize_uri(url)).strip()
        self._metadata = loads(self.http_get(__url))

        params = [
            'device_id=3',
            'manga_id={}'.format(self.re.search(r'/chapter/(\d+)', ch).group(1)),
            'page={page}',
        ]
        url = '{}/manga/get_manga_url?'.format(self.domain) + '&'.join(params)
        info('Chapter url: %s' % url)
        if self.__has_auth:
            params.append('client_login=true')
            info('Logged client!')
        else:
            info('Anon client!')

        self.http().proxies = _proxies.copy()

        return [url.format(page=i) for i in range(250)]  # fixme: max 250 images per chapter

    def get_cover(self):
        self._cover_from_content('.o_hero-media')

    def prepare_cookies(self):
        self._params['max_threads'] = 2
        self.http()._download = self.viz_downloader

        self.http().mute = True
        self.cookie_file = path_join(get_util_home_path(), 'cookies_viz_com.dat')
        cookies = self.load_cookies()
        content = self.http().requests(self.get_url(), cookies=cookies)
        cookies.update(content.cookies.get_dict())
        self.__cookies = cookies

        if not self.has_auth():
            self.auth()
            if not self.has_auth():
                self.log('Warning! Login/password incorrect?\nTry to get free chapters...', file=stderr)
                self.log('Warning! This site worked from USA and Japan! Check your location', file=stderr)
                unlink(self.cookie_file)
                return

        self.save_cookies(self.__cookies)
        self.http().cookies = self.__cookies

    def auth(self):
        token = self.get_token()

        name = self.arg('login') or self.quest([], 'Request login on viz.com')
        password = self.arg('password') or self.quest_password('Request password on viz.com\n')

        if len(name) == 0 or len(password) == 0:
            return

        req = self.http().requests(
            '{}/account/try_login'.format(self.domain),
            method='post', cookies=self.__cookies, data={
                'login': name,
                'pass': password,
                'rem_user': 1,
                'authenticity_token': token,
            })

        if req.status_code >= 400:
            self.log('Login/password error')
            exit(1)

        self.__cookies = req.cookies.get_dict()

        try:
            remember = self.json.loads(req.text)
            self.__cookies['remember_token'] = remember.get(
                'trust_user_id_token_web',
                remember.get('remember_token', '')
            )
        except ValueError:
            error('Remember error!')
            error('Please, report this error {}{}'.format(
                meta.repo_url, '/issues/new?template=bug_report.md'
            ))

    def save_cookies(self, cookies: dict):
        with open(self.cookie_file, 'w') as w:
            w.write(self.json.dumps(cookies))

    def load_cookies(self):
        if is_file(self.cookie_file):
            try:
                with open(self.cookie_file, 'r') as r:
                    return self.json.loads(r.read())
            except ValueError:
                unlink(self.cookie_file)
        return {}

    def get_token(self):
        auth_token_url = '{}/account/refresh_login_links'.format(self.domain)
        auth_token = self.http_get(auth_token_url, cookies=self.__cookies)
        token = self.re.search(r'AUTH_TOKEN\s*=\s*"(.+?)"', auth_token)
        return token.group(1)

    def has_auth(self):
        content = self.http_get('{}/account/refresh_login_links'.format(self.domain), cookies=self.__cookies)
        parser = self.document_fromstring(content)
        profile = parser.cssselect('.o_profile-link')
        success = len(profile) > 0
        self.__has_auth = success
        if success:
            # self.log('Login as {}'.format(profile[0].text))
            self.log('Login success')
        return success

    @staticmethod
    def has_chapters(parser):
        return len(parser.cssselect('.o_chapter-container')) > 0

    def before_download_file(self, idx, url):
        _path, idx, _url = super().before_download_file(idx, url)
        if not self._continue:
            return None, None, None
        info('\nSave file: {}'.format(idx))
        info('File url: {}'.format(_url))

        info('File params:\n PATH: {}\n IDX: {}\n URL: {}'.format(_path, idx, _url))

        self.http().cookies['chapter-series-5-follow-modal'] = time.strftime('%Y-%m-%d')

        __url = request(
            'get',
            self.http().normalize_uri(_url),
            headers={
                'X-Requested-With': 'XMLHttpRequest',
                'Referer': 'https://www.viz.com',
                'User-Agent': self.http().user_agent,
            },
            cookies=cookiejar_from_dict(self.http().cookies),
            allow_redirects=True
        ).text.strip()

        if __url.find('http') != 0:
            if self._continue:
                error('\nURL is wrong: \n {}\n'.format(__url))
            self._continue = False
            return None, None, None
        return _path, idx, __url

    def after_file_save(self, _path, idx):
        if file_size(_path) < 100:
            if self._continue:
                info('File not found. Stop for this chapter')
            self._continue = False
            is_file(_path) and unlink(_path)
            return None, None

        _path, arc_name = super().after_file_save(_path, idx)

        ref = solve(_path, self._metadata)

        self._wait_after_file()

        if ref is not None:
            solved_path = _path + '-solved.jpeg'
            ref.save(solved_path)
            return solved_path, 'solved{}.jpeg'.format(idx)
        return _path, arc_name


main = VizCom



try:
    import cloudscraper as cfscrape
except ModuleNotFoundError:
    import cfscrape

# from logging import error
from pathlib import Path
from urllib.parse import urlparse, ParseResult
import json
from typing import Optional, Tuple
from ..fs import get_util_home_path
from time import time


CF_TIMEOUT = 3600 * 24 * 7  # 7 days


def cf_scrape(url):  # pragma: no cover
    cf = cf_load(url)

    if cf is not None:
        return cf

    scraper = cfscrape.create_scraper()
    try:
        cookies, ua = scraper.get_tokens(url)

        cf_save(url, cookies, ua)

        return cookies, ua
    except Exception as e:
        raise RuntimeError('cloudflare module not supported')
        # error(e)
        # raise e


def __domain(url: str) -> str:
    parsed = urlparse(url, 'https')  # type: ParseResult
    return parsed.hostname


def __cookies_location(url: str) -> Path:
    domain = __domain(url)
    path = get_util_home_path()
    return Path(path).joinpath('cf_cookies', '{}.dat'.format(domain))


def cf_save(url, cookies, ua):
    path = __cookies_location(url)
    path.parent.mkdir(exist_ok=True, parents=True)

    path.write_text(json.dumps({
        'cookies': cookies,
        'ua': ua,
        'time': int(time()),
    }))


def cf_load(url) -> Optional[Tuple[dict, str, int]]:
    try:
        path = __cookies_location(url)
        data = json.loads(path.read_text())

        if time() - data['time'] > CF_TIMEOUT:
            return None

        return data['cookies'], data['ua'], data['time']
    except (FileNotFoundError, NameError, ValueError, TypeError):
        return None



from manga_py.provider import Provider
from .helpers.std import Std


class HentaiImageCom(Provider, Std):

    def get_archive_name(self) -> str:
        return 'archive'

    def get_chapter_index(self) -> str:
        return '0'

    def get_content(self):
        return self._get_content('{}/image/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/image/([^/]+)')

    def get_chapters(self):
        return [b'']

    def _pages(self, parser):
        pages = parser.cssselect('#paginator')
        if pages:
            href = pages[0].cssselect('span > a')[-1].get('href')
            page = self.re.search(r'/page/(\d+)', href)
            return range(2, int(page.group(1)) + 1)
        return []

    def get_files(self):
        parser = self.document_fromstring(self.content)
        pages = self._pages(parser)
        selector = '#display_image_detail div > a > img'
        images = self._images_helper(parser, selector)
        for i in pages:
            content = self._get_content('{}/image/{}/page/%d' % i)
            parser = self.document_fromstring(content)
            images += self._images_helper(parser, selector)
        return images

    def get_cover(self) -> str:
        pass

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.get_url()


main = HentaiImageCom



from .rawdevart_com_old import RawDevArtComOld


class Manga41Com(RawDevArtComOld):
    _chapter_selector = r'/chapter-(\d+(?:\.\d+)?)'

    def get_files(self):
        content = self.http_get(self.chapter)
        images = self.re.search(r'chapter_preloaded_images.+?(\[.+\])', content).group(1)
        return self.json.loads(images)


main = Manga41Com



from .base_lib import BaseLib
from logging import error


class KissMangaComCrypt(BaseLib):

    def decrypt(self, iv, key, data):
        iv = self.encode_hex(iv)
        key = self.to_sha_256(key)
        data = self.base64decode(data)

        try:
            return self.decrypt_aes(iv, key, data)
        except Exception as e:
            error(e)
            return False



from manga_py.provider import Provider
from .helpers.std import Std


class MangaParkOrg(Provider, Std):
    __url = None

    def get_chapter_index(self) -> str:
        return self.chapter[0].replace('.', '-')

    def get_content(self):
        return self.http_get(self.__url)

    def get_manga_name(self) -> str:
        title = self.html_fromstring(self.get_url(), 'h3 > a, h4 > a', 0)
        self.__url = self.http().normalize_uri(title.get('href'))
        return title.text_content().strip()

    def _print_variants(self, variants):
        self.log('Please, select lang. (empty for all langs)')
        for n, i in enumerate(variants):
            lng = i.cssselect('.card-header a')[0].text_content()
            self.log('\n%d: ' % (n + 1) + lng, end='')

    def _answer(self, max_digit):
        while True:
            answer = self.quest([], 'Answer (digit): ')
            if len(answer) > 0 and (int(answer) > max_digit or int(answer) <= 0):
                self.log('Wrong answer! Try one more.')
            else:
                return answer

    def _auto_language(self, variants):
        language = self.arg('language')

        if language is not None:

            for n, variant in enumerate(variants):
                text = variant.cssselect('.flag + a.ml-1')[0].text_content()  # type: str

                if ~text.find('[' + language + ']'):
                    return n + 1

        return None

    def _auto_translator(self, variants):
        translator = self.arg('translator')

        if translator is not None:

            for n, variant in enumerate(variants):
                text = variant.cssselect('.flag + a.ml-1')[0].text_content()  # type: str

                lng = text.find(']')
                if ~text.find(translator, lng):
                    return n + 1

        return None

    def _auto_rules(self, variants, default):
        auto_lang = self._auto_language(variants)
        auto_translator = self._auto_translator(variants)

        if auto_lang is not None:
            return auto_lang

        if auto_translator is not None:
            return auto_translator

        if len(variants) > 1:
            self._print_variants(variants)
            return self._answer(len(variants))

        return default

    def get_chapters(self):
        # multi-lang!
        variants = self._elements('div.card')
        answer = self._auto_rules(variants, '1')

        if len(variants) > 1 and not len(answer):
            parser = self.document_fromstring(self.content)
        else:
            parser = variants[int(answer) - 1]

        items = parser.cssselect('.card-body i + a')
        result = []
        re = self.re.compile(r'[Cc]h\.(\d+(?:\.\d+)?)')
        n = self.http().normalize_uri
        for i in items:
            text = i.text_content()
            result.append((
                re.search(text).group(1),
                n(i.get('href')),
            ))
        return result

    def get_files(self):
        re = self.re.compile(r'images\s*=\s*(\[.+\]);')
        content = self.http_get(self.chapter[1])
        items = self.json.loads(re.search(content).group(1))
        return items

    def get_cover(self) -> str:
        return self._cover_from_content('.order-0 > img')

    def book_meta(self) -> dict:
        pass

    def chapter_for_json(self) -> str:
        return self.chapter[1]

    def prepare_cookies(self):
        self.http().cookies['set'] = 'theme=1&h=1&img_load=4&img_zoom=5&img_tool=4&twin_m=0&twin_c=0&manga_a_warn=1&history=1&timezone=14'


main = MangaParkOrg



from manga_py.provider import Provider
from .helpers.std import Std


class MangaParkMe(Provider, Std):

    def get_chapter_index(self) -> str:
        selector = r'/manga/[^/]+/[^/]+(?:/v(\d+))?/c(\d+[^/]*)'
        idx = self.re.search(selector, self.chapter).groups()
        if idx[0] is None:
            return '0-' + idx[1]
        return '-'.join(idx)

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        return self._elements('#list a.ch')

    def get_files(self):
        chapter = self.re.search(r'^(.+/c\d+)', self.chapter).group(1)
        content = self.http_get(chapter)
        data = self.re.search(r'var\s_load_page\s?=\s?(\[.+\])', content)
        if not data:
            return []
        json = self.json.loads(data.group(1))
        return [i['u'] for i in json]

    def get_cover(self):
        return self._cover_from_content('.cover img')


main = MangaParkMe



from lxml.html import document_fromstring
# from purifier.purifier import HTMLPurifier


class Static:

    # @staticmethod
    # def _clear_html(body):
    #     purifier = HTMLPurifier({
    #         'div': ['*'], 'span': ['*'],
    #         'img': ['*'], 'a': ['*'],
    #         'h1': ['*'], 'h2': ['*'],
    #         'h3': ['*'], 'h4': ['*'],
    #         'h5': ['*'], 'h6': ['*'],
    #     })
    #     return purifier.feed(body)

    @staticmethod
    def document_fromstring(body, selector: str = None, idx: int = None):  # pragma: no cover
        result = document_fromstring(body)  # todo
        if isinstance(selector, str):
            result = result.cssselect(selector)
        if isinstance(idx, int):
            result = result[idx]
        return result

    @staticmethod
    def _set_if_not_none(var, key, value):  # pragma: no cover
        if value is not None:
            var[key] = value

    @staticmethod
    def __test_ascii(i):
        o = ord(i)
        _ = 39 < o < 127
        _ = _ and o not in [42, 47, 92, 94]
        return _ or o > 161

    @staticmethod
    def remove_not_ascii(value):
        return "".join(i for i in value if i == '_' or Static.__test_ascii(i))



from .taadd_com import TaaddCom
from time import sleep


class TenMangaCom(TaaddCom):
    __manga_url = 'https://www.gardenmanage.com/c/tenmanga/'
    _name_selector = '.read-page a[href*="/book/"]'
    _pages_selector = '.sl-page'
    _chapters_selector = '.chapter-name.long > a'
    img_selector = '.pic_box .manga_pic'

    def get_files(self):
        try:
            content = self.http_get(self.chapter)
        except ConnectionError:
            sleep(.2)
            content = self.http_get(self.chapter)

        images = self.re.search(r'all_imgs_url.+?\[((?:\s*"http.+?",?)+)', content)

        if not images:
            return []

        images = self.re.sub(r'["\s*]', '', images.group(1))
        images = images.rstrip(',').split(',')
        return images


main = TenMangaCom



from manga_py.provider import Provider
from .helpers.std import Std


class _Template(Provider, Std):
    def get_archive_name(self) -> str:
        """
        Allows you to overload name generation. Method may be missing
        """

    def get_chapter_index(self) -> str:
        """
        Any string separated by "-"
        Example: "1-3"
        """

    def get_content(self):
        """
        Called second. Returns the initial content to parse (just a cache)
        Can be obtained using self.content
        """

    def get_manga_name(self) -> str:
        """
        String for the name of the manga directory
        """

    def get_chapters(self):
        """
        Should return an array of data (you can return the result of the self._elements ('css selector') method)
        """
        # return self._elements('a.chapter')
        return []

    def get_files(self):
        """
        Should return an array of strings (url of images)
        """
        return []

    def get_cover(self) -> str:
        """Not used now"""
        # return self._cover_from_content('.cover img')

    def book_meta(self) -> dict:
        """
        Not used now
        :see http://acbf.wikia.com/wiki/Meta-data_Section_Definition
        return {
            'author': str,
            'title': str,
            'annotation': str,
            'keywords': str,
            'cover': str,
            'rating': str,
          }
        """

    def chapter_for_json(self) -> str:
        """
        overload std param, if need
        If present, overloads getting chapter name for json dump
        Should return a string
        """
        # return self.chapter
        pass


main = _Template



from manga_py.provider import Provider
from .helpers.std import Std


class WestMangaInfo(Provider, Std):
    _chapter_re = r'\.\w{2,7}/[^/]+-(\d+(?:-\d+)?)'

    def get_chapter_index(self) -> str:
        re = self.re.compile(self._chapter_re)
        return re.search(self.chapter).group(1)

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        url = self.get_url()
        if ~url.find('/manga/'):
            return self._get_name('/manga/([^/]+)')
        url = self.html_fromstring(url, '.allc a', 0).get('href')
        self._params['url'] = self.http().normalize_uri(url)
        return self.get_manga_name()

    def get_chapters(self):
        return self._elements('span.leftoff > a')

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        return self._images_helper(parser, '.lexot img')

    def get_cover(self) -> str:
        return self._cover_from_content('.naru img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = WestMangaInfo



from manga_py.provider import Provider
from .helpers.std import Std


class ZeroScansCom(Provider, Std):
    _key = '/comics/'

    def get_chapter_index(self) -> str:
        return self.re.search(
            r'%s[^/]+/(\d+/\d+)' % self._key,
            self.chapter
        ).group(1).replace('/', '-')

    def get_content(self):
        name = self._get_name(r'%s([^/]+)' % self._key)
        return self.http_get('%s%s%s/' % (
            self.domain,
            self._key,
            name
        ))

    def get_manga_name(self) -> str:
        return self._get_name(r'%s\d+-([^/]+)' % self._key)

    def get_chapters(self):
        return self._elements('.list .list-item a.text-color')

    def get_files(self):
        content = self.http_get(self.chapter)
        raw_images = self.re.search(
            r'chapterPages\s?=\s?(\[.+?\])',
            content
        ).group(1)
        images = self.json.loads(raw_images)

        n = self.http().normalize_uri

        return [n(i) for i in images]

    def get_cover(self) -> str:
        image = self._elements('.media img.media-content')
        if len(image):
            return self.parse_background(image)
        return ''


main = ZeroScansCom



from sys import stderr

from manga_py.provider import Provider
from .helpers.std import Std


class MangAs(Provider, Std):

    def get_chapter_index(self) -> str:
        idx = self.re.search('/manga/[^/]+/([^/]+)', self.chapter).group(1)
        return idx.replace('.', '-')

    def get_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        return self._elements('.chapter-title-rtl > a')

    def get_files(self):
        content = self.http_get(self.chapter)
        self.http().referer = self.chapter
        items = self.re.search(r'var\s+pages\s*=\s*(\[.+\])', content)
        if not items:
            self.log('Images not found!', file=stderr)
            return []
        n = self.http().normalize_uri
        items = self.json.loads(items.group(1))
        return [n(i.get('page_image')) for i in items]

    def prepare_cookies(self):
        self._base_cookies(self.get_url())

    def get_cover(self) -> str:
        return self._cover_from_content('.boxed > img.img-responsive')

    def book_meta(self) -> dict:
        pass


main = MangAs




def _args_image(args_parser):  # pragma: no cover
    args = args_parser.add_argument_group('Image options')

    args.add_argument(
        '-E',
        '--not-change-files-extension',
        action='store_true',
        help=(
            'Save downloaded files to archive "as is".'
        )
    )

    args.add_argument(
        '-W',
        '--no-webp',
        action='store_true',
        help=(
            'Convert `*.webp` images to `*.jpg` format.'
        )
    )
